module wrf2gem_subs
  use netcdf, only: nf90_open, nf90_inq_dimid, nf90_inquire_dimension,  &
                    nf90_get_att, nf90_inq_varid, nf90_get_var, nf90_close,  &
                    nf90_inquire_variable
  use netcdf, only: NF90_NoWrite, NF90_NoErr, NF90_Global, NF90_StrError,  &
                    NF90_Max_Var_Dims
  implicit none
  save
  private

  public :: get_out_fields, get_num_files, open_wrf_file, get_grid_info,  &
            update_grid_info, get_times, close_wrf_file, handle_err,  &
            mem_error, read_values, freq_divisible
  public :: NLUnit, NLFName

  integer,           parameter :: NLUnit = 19, MaxOuts = 200
  character(len=10), parameter :: NLFName = "convert.nl"

  integer,           dimension(MaxOuts), target :: outFields
  character(len=80), dimension(MaxOuts), target :: fName

  interface read_values
     module procedure rv0d
     module procedure rv1d
     module procedure rv2d
     module procedure rv3d
     module procedure rv4d
  end interface


contains  ! ===================================================================

  
  ! get_out_fields reads the namelist to determine what variables should be
  ! output to the GEMPAK file.  It returns the number of variables to output
  ! (num) and a pointer to an array of variable IDs (ptr).  If there are no
  ! variables to output, the status of ptr does not change.  The subroutine
  ! also determines how pressure interpolation should be handled, giving
  ! values for the bottom and top pressure levels (pb, pt) and the pressure
  ! interval (dp).
  subroutine get_out_fields(num, ptr, pb, pt, dp)
    integer, intent(out) :: num
    integer, dimension(:), pointer :: ptr
    integer, intent(out) :: pb, pt, dp

    integer :: status
    logical :: nlExist

    namelist /NUMOUTS/ num
    namelist /VARS/ outFields, pb, pt, dp

    ! The default if there is no namelist information is to output nothing.
    num = 0

    ! Defaults for pressure level interpolation information
    pb = 1000; pt = 200; dp = 50

    ! Check for namelist existence, then read it.
    inquire (file=NLFName, exist=nlExist)
    if (nlExist) then
       open (unit=NLUnit, file=NLFName, iostat=status)
       if (status /= 0) stop "There is a problem opening the namelist file."
       read (unit=NLUnit, nml=NUMOUTS, iostat=status)
       if (status /= 0) stop "There is a problem reading NUMOUTS from the &
            &namelist file."
       
       if (num > MaxOuts) stop "There are too many outputs!"

       read (unit=NLUnit, nml=VARS, iostat=status)
       if (status /= 0) stop "There is a problem reading VARS from the &
            &namelist file."
       close (unit=NLUnit)

       if (num > 0) ptr => outFields(1:num)
    end if
  end subroutine get_out_fields

  ! ===========================================================================
  
  ! get_num_files reads the namelist to determine what WRF files should be
  ! read for conversion.  It returns the number of files to convert (numFiles)
  ! and a pointer to an array of filenames (point).  If there are no files to
  ! convert, the status of point does not change.
  subroutine get_num_files(numFiles, point)
    integer, intent(out) :: numFiles
    character(len=80), dimension(:), pointer :: point
    
    integer :: status
    logical :: nlExist

    namelist /OUTFILES/ numFiles
    namelist /WRFFILE/ fName

    ! The default if there is no namelist information is to output nothing.
    numFiles = 0

    ! Check for namelist existence, then read it.
    inquire (file=NLFName, exist=nlExist)
    if (nlExist) then
       open (unit=NLUnit, file=NLFName, iostat=status)
       if (status /= 0) stop "There is a problem opening the namelist file."
       read (unit=NLUnit, nml=OUTFILES, iostat=status)
       if (status /= 0) stop "There is a problem reading OUTFILES from the &
            &namelist file."

       if (numFiles > MaxOuts) stop "There are too many files to process!"

       read (unit=NLUnit, nml=WRFFILE, iostat=status)
       if (status /= 0) stop "There is a problem reading WRFFILE from the &
            &namelist file."
       close (unit=NLUnit)

       if (numFiles > 0) point => fName(1:numFiles)
    end if
  end subroutine get_num_files

  ! ===========================================================================

  ! open_wrf_file opens the WRF history file (fil) and returns its handle (id).
  subroutine open_wrf_file(fil, id)
    character(len=80), intent(in)  :: fil
    integer,           intent(out) :: id
    
    integer :: status
    
    status = nf90_open(fil, NF90_NoWrite, id)
    call handle_err(status)    
  end subroutine open_wrf_file

  ! ===========================================================================

  ! get_grid_info gathers the following info from the WRF history file (ncid):
  ! o Model grid extent in x, y, and z directions (nx, ny, nz)
  ! o Number of times in the file (nt)
  ! o Map projection type (proj)
  ! o Latitude/longitude in lower-left and upper-right grid corners (lat, lon)
  ! o Projection angle info (ang)
  subroutine get_grid_info(ncid, nx, ny, nz, nt, proj, lat, lon, ang)
    integer,            intent(in)  :: ncid
    integer,            intent(out) :: nx, ny, nz, nt, proj
    real, dimension(2), intent(out) :: lat, lon
    real, dimension(3), intent(out) :: ang
    
    character(len=11), dimension(4), parameter :: DimName = (/ "west_east  ", &
                                 "south_north", "bottom_top ", "Time       " /)
    integer, dimension(4) :: temp
    integer               :: i, varid, status

    do i = 1, 4
       status = nf90_inq_dimid(ncid, trim(DimName(i)), varid)
       call handle_err(status)  
       status = nf90_inquire_dimension(ncid, varid, len=temp(i))
       call handle_err(status)
    end do
    nx = temp(1)
    ny = temp(2)
    nz = temp(3)
    nt = temp(4)
        
    status = nf90_get_att(ncid, NF90_Global, "MAP_PROJ", proj)
    call handle_err(status) 
    
    status = nf90_inq_varid(ncid, "XLONG", varid)
    call handle_err(status)
    status = nf90_get_var(ncid, varid, lon(1), start = (/ 1, 1, 1 /))
    call handle_err(status)
    status = nf90_get_var(ncid, varid, lon(2), start = (/ nx, ny, 1 /))
    call handle_err(status)
    
    status = nf90_inq_varid(ncid, "XLAT", varid)
    call handle_err(status)
    status = nf90_get_var(ncid, varid, lat(1), start = (/ 1, 1, 1 /))
    call handle_err(status)
    status = nf90_get_var(ncid, varid, lat(2), start = (/ nx, ny, 1 /))
    call handle_err(status)
    
    select case (proj)
    case (1) ! LCC
       status = nf90_get_att(ncid, NF90_Global, "TRUELAT1", ang(1))
       call handle_err(status) 
       status = nf90_get_att(ncid, NF90_Global, "CEN_LON", ang(2))
       call handle_err(status) 
       status = nf90_get_att(ncid, NF90_Global, "TRUELAT2", ang(3))
       call handle_err(status)
    case (3) ! Mercator
       ang = 0
    case default
       write (*,"(a,i2)") "proj = ", proj
       print *, "This map projection has not yet been implemented."
       stop
    end select
  end subroutine get_grid_info

  ! ===========================================================================
  
  ! update_grid_info compares grid size and prejection settings from the WRF
  ! history file (ncid) with the following previously saved settings:
  ! o Model grid extent in x, y, and z directions (nx, ny, nz)
  ! o Map projection type (proj)
  ! o Latitude/longitude in lower-left and upper-right grid corners (lat, lon)
  ! o Projection angle info (ang)
  ! If there is not an exact match, ok is set to .false.; otherwise, ok is
  ! .true. The number of times in WRF file (nt) is also provided as output.
  subroutine update_grid_info(ncid, nx, ny, nz, proj, lat, lon, ang, nt, ok)
    integer,            intent(in)  :: ncid, nx, ny, nz, proj
    real, dimension(2), intent(in)  :: lat, lon
    real, dimension(3), intent(in)  :: ang
    integer,            intent(out) :: nt
    logical,            intent(out) :: ok
    
    character(len=11), dimension(4), parameter :: DimName = (/ "west_east  ", &
                                 "south_north", "bottom_top ", "Time       " /)
    real,    dimension(3) :: angl
    real,    dimension(2) :: lonl, latl
    integer, dimension(4) :: temp
    integer               :: i, varid, status, nxl, nyl, nzl, projl

    do i = 1, 4
       status = nf90_inq_dimid(ncid, trim(DimName(i)), varid)
       call handle_err(status)  
       status = nf90_inquire_dimension(ncid, varid, len=temp(i))
       call handle_err(status)
    end do
    nxl = temp(1)
    nyl = temp(2)
    nzl = temp(3)
    nt = temp(4)
        
    status = nf90_get_att(ncid, NF90_Global, "MAP_PROJ", projl)
    call handle_err(status) 
    
    status = nf90_inq_varid(ncid, "XLONG", varid)
    call handle_err(status)
    status = nf90_get_var(ncid, varid, lonl(1), start = (/ 1, 1, 1 /))
    call handle_err(status)
    status = nf90_get_var(ncid, varid, lonl(2), start = (/ nxl, nyl, 1 /))
    call handle_err(status)
    
    status = nf90_inq_varid(ncid, "XLAT", varid)
    call handle_err(status)
    status = nf90_get_var(ncid, varid, latl(1), start = (/ 1, 1, 1 /))
    call handle_err(status)
    status = nf90_get_var(ncid, varid, latl(2), start = (/ nxl, nyl, 1 /))
    call handle_err(status)
    
    select case (projl)
    case (1) ! LCC
       status = nf90_get_att(ncid, NF90_Global, "TRUELAT1", angl(1))
       call handle_err(status) 
       status = nf90_get_att(ncid, NF90_Global, "CEN_LON", angl(2))
       call handle_err(status) 
       status = nf90_get_att(ncid, NF90_Global, "TRUELAT2", angl(3))
       call handle_err(status)
    case (3) ! Mercator
       angl = 0
    case default
       write (*,"(a,i2)") "proj = ", proj
       print *, "This map projection has not yet been implemented."
       stop
    end select

    ! Check for exact match
    ok = (nxl == nx .and. nyl == ny .and. nzl == nz .and.  &
         projl == proj .and. all(latl == lat) .and. all(lonl == lon) .and.  &
         all(angl == ang))
  end subroutine update_grid_info

  ! ===========================================================================

  ! get_times uses information from the WRF history file (id) and the namelist 
  ! to create an appropriate array of output times in GEMPAK form (times).  In
  ! addition, the number of outputs per 12 hours (timesPer12hr) is provided for
  ! output.  This subroutine assumes the history file contains either output
  ! every n hours, where n is some positive integer, or output for one time.
  subroutine get_times(id, times, timesPer12hr)
    integer,                        intent(in)  :: id
    character(len=*), dimension(:), intent(out) :: times
    integer,                        intent(out) :: timesPer12hr

    character(len=19), dimension(size(times,1)) :: wrfTimes
    integer :: numTimes, varid, status, initYear, initMonth, initDate,  &
               initHour, origInitYear, origInitMonth, origInitDate,  &
               origInitHour, ihr1, ihr2, interval, i
    character(len=3) :: fhr
    character(len=2) :: chr1, chr2
    logical          :: nlExist

    namelist /INITTIME/ initYear, initMonth, initDate, initHour

    ! Initialization
    numTimes = size(times,1)
    
    ! Read WRF history file times
    ! The format for these times is yyyy-mm-dd_hh:mm:ss
    status = nf90_inq_varid(id, "Times", varid)
    call handle_err(status)
    status = nf90_get_var(id, varid, wrfTimes)
    call handle_err(status)

    ! Determine a default initial time from the first time in the history file
    initYear = char_to_int(wrfTimes(1)(1:4));  origInitYear  = initYear
    initMonth = char_to_int(wrfTimes(1)(6:7)); origInitMonth = initMonth
    initDate = char_to_int(wrfTimes(1)(9:10)); origInitDate  = initDate
    chr1 = wrfTimes(1)(12:13)
    initHour = char_to_int(chr1);              origInitHour  = initHour

    ! Calculate interval between outputs in hours, and number of outputs per 12
    ! hours.
    if (numTimes > 1) then
       chr2 = wrfTimes(2)(12:13)  ! Hour at second time
       read (unit=chr1, fmt=*) ihr1
       read (unit=chr2, fmt=*) ihr2
       interval = ihr2 - ihr1
       timesPer12hr = 12 / interval
    else
       timesPer12hr = 1
    end if

    ! Use values from namelist to override defaults
    inquire (file=NLFName, exist=nlExist)
    if (nlExist) then
       open (unit=NLUnit, file=NLFName, iostat=status)
       if (status /= 0) stop "There is a problem opening the namelist file."
       read (unit=NLUnit, nml=INITTIME, iostat=status)
       if (status /= 0) stop "There is a problem reading INITTIME from the &
            &namelist file."
       close (unit=NLUnit)
    end if
    
    ! Calculate forecast hour to which the first time in history file
    ! corresponds
    ihr1 = init_fhr(origInitYear, origInitMonth, origInitDate, origInitHour,  &
         initYear, initMonth, initDate, initHour)
    if (ihr1 < 0) then
       print *, "Values in the namelist INITTIME are invalid. Reverting to&
            & defaults."
       initYear = origInitYear
       initMonth = origInitMonth
       initDate = origInitDate
       initHour = origInitHour
       ihr1 = 0
    end if

    ! Form GEMPAK times for all of the times in the history file
    ! GEMPAK format is yymmdd/hhmmFfff
    do i = 1, numTimes
       times(i)(1:2) = int_to_char2(initYear)
       times(i)(3:4) = int_to_char2(initMonth)
       times(i)(5:6) = int_to_char2(initDate)
       times(i)(7:7) = "/"
       times(i)(8:9) = int_to_char2(initHour)
       times(i)(10:11) = wrfTimes(1)(15:16)
       times(i)(12:12) = "F"

       write (unit=fhr, fmt="(i3)") ihr1
       select case (ihr1)
       case (0:9)
          fhr(1:2) = "00"
       case (10:99)
          fhr(1:1) = "0"
       end select

       times(i)(13:15) = fhr
       ihr1 = ihr1 + interval
    end do
  end subroutine get_times

  ! ===========================================================================

  ! close_wrf_file closes the WRF history file specified by id.
  subroutine close_wrf_file(id)
    integer, intent(in) :: id
    
    integer :: status
    
    status = nf90_close(id)
    call handle_err(status)
  end subroutine close_wrf_file
    
  ! ===========================================================================

  ! handle_err examines the result code from a netCDF function. If necessary,
  ! it then prints a netCDF error message and aborts the program.
  subroutine handle_err(status)
    integer, intent(in) :: status
    
    if (status /= NF90_NoErr) then
       print *, trim(NF90_StrError(status))
       print *, "Stopped"
       stop
    end if
  end subroutine handle_err

  ! ===========================================================================
  
  ! mem_error examines the result code from a memory allocation/deallocation.
  ! If necessary, it then prints an error message and aborts the program.
  subroutine mem_error(status, id, string)
    integer, intent(in) :: status, id
    character(len=*), intent(in) :: string

    if (status /= 0) then
       select case (id)
       case (1)
          print *, "Memory allocation error in ", string
       case (2)
          print *, "Memory deallocation error in ", string
       case default
          print *, "Memory error in ", string
       end select
       
       stop
    end if
  end subroutine mem_error

  ! ===========================================================================

  ! rv0d supplies a scalar containing the variable specified by varName from
  ! the netCDF file associated with id.
  subroutine rv0d(id, varName, value)
    integer,          intent(in)  :: id
    character(len=*), intent(in)  :: varName
    real,             intent(out) :: value
    
    integer, dimension(NF90_Max_Var_Dims) :: dimIDs
    integer :: varid, status, nDims, ncSize
    
    ! Get array extents from the netCDF file
    status = nf90_inq_varid(id, varName, varid)
    call handle_err(status)
    status = nf90_inquire_variable(id, varid, ndims = nDims, dimids = dimIDs)
    call handle_err(status)
    status = nf90_inquire_dimension(id, dimIDs(1), len = ncSize)
    call handle_err(status)
    
    ! Make sure array extents match!
    if (ncSize /= 1 .or. nDims /= 0) then
       print *, "Array dimensions are incorrect in read_values."
       stop
    end if
    
    ! Read the variable from the netCDF file
    status = nf90_get_var(id, varid, value)
    call handle_err(status)
  end subroutine rv0d

  ! ...........................................................................

  ! rv1d supplies a 1D array containing the variable specified by varName from
  ! the netCDF file associated with id.
  subroutine rv1d(id, varName, values)
    integer,            intent(in)  :: id
    character(len=*),   intent(in)  :: varName
    real, dimension(:), intent(out) :: values
    
    integer, parameter :: NumDims = 1
    integer, dimension(NF90_Max_Var_Dims) :: dimIDs
    integer, dimension(NumDims)           :: sizes, ncSizes
    integer :: i, varid, status, nDims
    
    ! Determine array extents that wrf2gem expects
    sizes = shape(values)
    
    ! Get array extents from the netCDF file
    status = nf90_inq_varid(id, varName, varid)
    call handle_err(status)
    status = nf90_inquire_variable(id, varid, ndims = nDims, dimids = dimIDs)
    call handle_err(status)
    do i = 1, NumDims
       status = nf90_inquire_dimension(id, dimIDs(i), len = ncSizes(i))
       call handle_err(status)
    end do
    
    ! Make sure array extents match!
    if (any(ncSizes /= sizes) .or. nDims /= NumDims) stop  &
         "Array dimensions are incorrect in read_values."
    
    ! Read the variable from the netCDF file
    status = nf90_get_var(id, varid, values)
    call handle_err(status)
  end subroutine rv1d

  ! ...........................................................................

  ! rv2d is identical to rv1d, but for a 2D array.
  subroutine rv2d(id, varName, values)
    integer,              intent(in)  :: id
    character(len=*),     intent(in)  :: varName
    real, dimension(:,:), intent(out) :: values
    
    integer, parameter :: NumDims = 2
    integer, dimension(NF90_Max_Var_Dims) :: dimIDs
    integer, dimension(NumDims)           :: sizes, ncSizes
    integer :: i, varid, status, nDims
    
    ! Determine array extents that wrf2gem expects
    sizes = shape(values)
    
    ! Get array extents from the netCDF file
    status = nf90_inq_varid(id, varName, varid)
    call handle_err(status)
    status = nf90_inquire_variable(id, varid, ndims = nDims, dimids = dimIDs)
    call handle_err(status)
    do i = 1, NumDims
       status = nf90_inquire_dimension(id, dimIDs(i), len = ncSizes(i))
       call handle_err(status)
    end do
    
    ! Make sure array extents match!
    if (any(ncSizes /= sizes) .or. nDims /= NumDims) stop  &
         "Array dimensions are incorrect in read_values."

    ! Read the variable from the netCDF file
    status = nf90_get_var(id, varid, values)
    call handle_err(status)
  end subroutine rv2d

  ! ...........................................................................
  
  ! rv3d is identical to rv2d, but for a 3D array.
  subroutine rv3d(id, varName, values)
    integer,                intent(in)  :: id
    character(len=*),       intent(in)  :: varName
    real, dimension(:,:,:), intent(out) :: values

    integer, parameter :: NumDims = 3
    integer, dimension(NF90_Max_Var_Dims) :: dimIDs
    integer, dimension(NumDims)           :: sizes, ncSizes
    integer :: i, varid, status, nDims

    ! Determine array extents that wrf2gem expects
    sizes = shape(values)
    
    ! Get array extents from the netCDF file
    status = nf90_inq_varid(id, varName, varid)
    call handle_err(status)
    status = nf90_inquire_variable(id, varid, ndims = nDims, dimids = dimIDs)
    call handle_err(status)
    do i = 1, NumDims
       status = nf90_inquire_dimension(id, dimIDs(i), len = ncSizes(i))
       call handle_err(status)
    end do

    ! Make sure array extents match!
    if (any(ncSizes /= sizes) .or. nDims /= NumDims) stop  &
         "Array dimensions are incorrect in read_values."

    ! Read the variable from the netCDF file
    status = nf90_get_var(id, varid, values)
    call handle_err(status)
  end subroutine rv3d

  ! ...........................................................................

  ! rv4d is identical to rv2d, but for a 4D array.
  subroutine rv4d(id, varName, values)
    integer,                  intent(in)  :: id
    character(len=*),         intent(in)  :: varName
    real, dimension(:,:,:,:), intent(out) :: values
    
    integer, parameter :: NumDims = 4
    integer, dimension(NF90_Max_Var_Dims) :: dimIDs   
    integer, dimension(NumDims)           :: sizes, ncsizes
    integer :: i, varid, status, nDims

    ! Determine array extents that wrf2gem expects
    sizes = shape(values)
    
    ! Get array extents from the netCDF file
    status = nf90_inq_varid(id, varName, varid)
    call handle_err(status)
    status = nf90_inquire_variable(id, varid, ndims = nDims, dimids = dimIDs)
    call handle_err(status)
    do i = 1, NumDims
       status = nf90_inquire_dimension(id, dimIDs(i), len = ncSizes(i))
       call handle_err(status)
    end do
    
    ! Make sure array extents match!
    if (any(ncSizes /= sizes) .or. nDims /= NumDims) stop  &
         "Array dimensions are incorrect in read_values."

    ! Read the variable from the netCDF file
    status = nf90_get_var(id, varid, values)
    call handle_err(status)
  end subroutine rv4d

  ! ===========================================================================
  
  ! Given that there are p outputs per q hours, can the outputs be subdivided
  ! into blocks of r hours? This function answers that question.
  ! Yes = .true., and No = .false.
  ! Negative values for p, q, or r will also give .false. as output.
  logical function freq_divisible(p, q, r)
    integer, intent(in) :: p, q, r
    
    integer :: gcf, pReduced, qReduced, outputTimestep
    
    ! Make sure p, q, and r are all positive
    if (p <= 0 .or. q <= 0 .or. r <= 0) then
       freq_divisible = .false.
       return
    end if
    
    ! Reduce the fraction p/q to lowest terms.
    gcf = greatest_common_factor(p, q)
    pReduced = p / gcf
    qReduced = q / gcf
    
    ! First test: If r is divisible by q, then certainly there is a subdivision
    ! into r; in fact, there are pr blocks.
    ! Second test: If p doesn't divide into q, then the output timestep is not
    ! an integer, which is not allowed if r is not divisible by q.
    ! Third test: If r is a multiple of the output timestep, we've found the 
    ! subdivision we're looking for.
    
    if (mod(r,qReduced) == 0) then
       freq_divisible = .true.
    else if (mod(qReduced,pReduced) /= 0) then
       freq_divisible = .false.
    else
       outputTimestep = qReduced / pReduced
       if (mod(r,outputTimestep) == 0) then
          freq_divisible = .true.
       else
          freq_divisible = .false.
       end if
    end if
  end function freq_divisible

  ! ===========================================================================

  ! greatest_common_factor uses the Euclidean algorithm to compute the GCF of
  ! integers a and b. It returns zero if both a and b are zero. Otherwise, it
  ! returns a positive integer.
  recursive function greatest_common_factor(a, b) result (gcf)
    integer, intent(in) :: a, b
    integer             :: gcf
    
    if (b == 0) then
       gcf = a
    else
       gcf = greatest_common_factor(b, mod(a,b))
    end if
  end function greatest_common_factor
  
  ! ===========================================================================

  ! This function converts the input character string (ch) to an integer.
  integer function char_to_int(ch)
    character(len=*), intent(in) :: ch

    character(len=*), parameter :: digs = "0123456789 -", frmt = "(i9)"
    integer :: chLen, i, index
    character(len=len(ch)) :: chl

    chLen = len(ch)
    chl = ch

    if (chLen < 1) then
       char_to_int = 0
       return
    end if

    ! Replace bad characters with spaces
    do i = 1, chLen
       if (scan(chl(i:i),digs) == 0) chl(i:i) = " "
    end do

    ! Make sure only spaces precede a minus sign
    index = scan(chl,"-")
    chl(1:index-1) = " "

    ! Make sure no additional minus signs occur
    do i = index+1, chLen
       if (chl(i:i) == "-") chl(i:i) = " "
    end do

    read (chl, frmt) char_to_int
  end function char_to_int

  ! ===========================================================================
  
  ! This function converts a positive integer (num) into a character of 
  ! length 2. If num < 0, it's absolute value is used. If num has more than
  ! 2 digits, it is truncated such that only the ones and tens digits convert.
  ! A leading zero is added if necessary.
  character(len=2) function int_to_char2(num)
    integer, intent(in) :: num

    integer :: numl

    numl = abs(num)
    if (numl > 99) numl = mod(numl,100)
    
    write(unit=int_to_char2, fmt="(i2)") numl
    if (numl < 10) int_to_char2(1:1) = "0"
  end function int_to_char2

  ! ===========================================================================

  ! init_fhr returns the number of hours elapsed from year1, month1, date1,
  ! hour1 to year2, month2, date2, hour2.  If the second time is before the
  ! first, -1 is returned.
  integer function init_fhr(year2, month2, date2, hour2, year1, month1,  &
       date1, hour1)
    integer, intent(in) :: year2, month2, date2, hour2, year1, month1,  &
                           date1, hour1
    
    integer, parameter :: MaxHours = 999
    integer :: year, month, date, hour, i

    ! Make sure the second time doesn't come before the first time
    if (year2 < year1 .or. year2 == year1 .and. month2 < month1 .or.  &
         year2 == year1 .and. month2 == month1 .and. date2 < date1 .or.  &
         year2 == year1 .and. month2 == month1 .and. date2 == date1 .and.  &
         hour2 < hour1) then
       init_fhr = -1
    else if (year2 == year1 .and. month2 == month1 .and.  &
         date2 == date1 .and. hour2 == hour1) then  ! Times match
       init_fhr = 0
    else
       year = year1; month = month1; date = date1; hour = hour1
       init_fhr = 0
       do i = 1, MaxHours  ! Progress forward in time hour by hour
          init_fhr = init_fhr + 1
          hour = hour + 1
          if (hour > 23) then  ! New day
             hour = 0
             date = date + 1
             if (date > days_in_month(month, year)) then  ! New month
                date = 1
                month = month + 1
                if (month > 12) then  ! New year
                   month = 1
                   year = year + 1
                end if
             end if
          end if
          
          ! Check to see if it is the same time
          if (year == year2 .and. month == month2 .and. date == date2 .and.  &
               hour == hour2) exit
       end do
    end if
  end function init_fhr

  ! ===========================================================================

  ! days_in_month returns the number of days month has given the year.
  integer function days_in_month(month, year)
    integer, intent(in) :: month, year

    select case (month)
    case (4,6,9,11)
       days_in_month = 30
    case (1,3,5,7,8,10,12)
       days_in_month = 31
    case (2)
       if (mod(year, 4) == 0 .and.  &
            (mod(year, 100) /= 0 .or. mod(year, 400) == 0)) then
          days_in_month = 29
       else
          days_in_month = 28
       end if
    case default
       stop "Invalid input to days_in_month!"
    end select
  end function days_in_month
end module wrf2gem_subs
